
import logging
import time
from StringIO import StringIO

from ensemble.environment.errors import EnvironmentsConfigError
from ensemble.control import setup_logging, main, setup_parser
from ensemble.lib.testing import TestCase

from .common import ControlToolTest


class ControlInitiailizationTest(ControlToolTest):

    # The EnvironmentTestBase will replace our $HOME, so that tests will
    # write properly to a temporary directory.
    def test_write_sample_config(self):
        """
        When ensemble-control is called without a valid environment
        configuration, it should write one down and raise an error to
        let the user know it should be edited.
        """
        try:
            main([])
        except EnvironmentsConfigError, error:
            self.assertTrue(error.sample_written)
        else:
            self.fail("EnvironmentsConfigError not raised")


class ControlCommandOptionTest(ControlToolTest):

    def test_command_suboption(self):
        """
        The argparser setup, invokes command module configure_subparser
        functions to allow the command to delineate additional cli
        options.
        """
        from ensemble.control import shutdown, bootstrap

        def configure_shutdown(subparser):
            subparser.add_argument("--opt1", default=1)

        def configure_bootstrap(subparser):
            subparser.add_argument("--opt2", default=2)

        self.patch(shutdown, "configure_subparser", configure_shutdown)
        self.patch(bootstrap, "configure_subparser", configure_bootstrap)

        parser = setup_parser()

        shutdown_opts = parser.parse_args(["shutdown"])
        bootstrap_opts = parser.parse_args(["bootstrap"])

        missing = object()
        self.assertEquals(shutdown_opts.opt1, 1)
        self.assertEquals(getattr(shutdown_opts, "opt2", missing), missing)
        self.assertEquals(bootstrap_opts.opt2, 2)
        self.assertEquals(getattr(bootstrap_opts, "opt1", missing), missing)


class ControlLoggingTest(TestCase):

    def test_logging_format(self):
        self.log = StringIO()
        setup_logging(False, stream=self.log)

        # ensure that we use gtm regardless of system settings
        logging.getLogger().handlers[0].formatter.converter = time.gmtime

        record = logging.makeLogRecord(
            {'created': 0, 'msecs': 0, 'levelno': logging.INFO})
        logging.getLogger().handle(record)
        self.assertEqual(
            self.log.getvalue(),
            '1970-01-01 00:00:00,000 Level None \n')

    def test_default_logging(self):
        """
        Default log-level is informational.
        """
        self.log = self.capture_logging()
        setup_logging(False)
        root = logging.getLogger()
        name = logging.getLevelName(root.getEffectiveLevel())
        self.assertEqual(name, 'INFO')

    def test_verbose_logging(self):
        """
        When verbose logging is enabled, the log level is set to debugging.
        """
        setup_logging(True)
        root = logging.getLogger()
        self.assertEqual(logging.getLevelName(root.level), 'DEBUG')
        custom = logging.getLogger('custom')
        self.assertEqual(custom.getEffectiveLevel(), root.getEffectiveLevel())

    def test_default_loggers(self):
        """
        Verify that the default loggers are bound when the logging
        system is started.
        """
        root = logging.getLogger()
        self.assertEqual(root.handlers, [])
        setup_logging()
        self.assertNotEqual(root.handlers, [])

    def tearDown(self):
        # remove the logging handlers we installed
        root = logging.getLogger()
        root.handlers = []
