import os
import random
import yaml
import uuid

from ensemble.lib.schema import SchemaError, KeyDict, Dict, String, Constant
from ensemble.errors import (
    FileNotFound, FileAlreadyExists, InvalidEnsembleHeaderValue)

from ensemble.environment.environment import Environment
from ensemble.environment.errors import EnvironmentsConfigError


DEFAULT_CONFIG_PATH = "~/.ensemble/environments.yaml"

SAMPLE_CONFIG = """\
ensemble: environments

environments:
  sample:
    type: ec2
    control-bucket: %(control-bucket)s
    admin-secret: %(admin-secret)s
"""


SCHEMA = KeyDict({
    "ensemble": Constant("environments"),
    "default": String(),
    "environments": Dict(String(), KeyDict({"type": String()})),
    },
    optional=["default"])


class EnvironmentsConfig(object):
    """An environment configuration, with one or more environments.
    """

    def __init__(self):
        self._config = None
        self._loaded_path = None

    def get_default_path(self):
        """Return the default environment configuration file path."""
        return os.path.expanduser(DEFAULT_CONFIG_PATH)

    def _get_path(self, path):
        if path is None:
            return self.get_default_path()
        return path

    def load(self, path=None):
        """Load an enviornment configuration file.

        @param path: An optional environment configuration file path.
            Defaults to ~/.ensemble/environments.yaml

        This method will call the C{parse()} method with the content
        of the loaded file.
        """
        path = self._get_path(path)
        if not os.path.isfile(path):
            raise FileNotFound(path)
        with open(path) as file:
            self.parse(file.read(), path)

    def parse(self, content, path=None):
        """Parse an enviornment configuration.

        @param content: The content to parse.
        @param path: An optional environment configuration file path, used
            when raising errors.

        @raise InvalidEnsembleHeaderValue: If 'ensemble:' is not set to
               'environments'.
        @raise EnvironmentsConfigError: On other issues.
        """
        config = yaml.load(content)
        found_value = config.get("ensemble")
        expected_value = "environments"
        if found_value != expected_value:
            raise InvalidEnsembleHeaderValue(path, expected_value, found_value)
        try:
            config = SCHEMA.coerce(config, [])
        except SchemaError, error:
            if path is None:
                path_info = ""
            else:
                path_info = " %s:" % (path,)
            raise EnvironmentsConfigError(
                "Environments configuration error:%s %s" %
                (path_info, str(error)))
        self._config = config
        self._loaded_path = path

    def get_names(self):
        """Return the names of environments available in the configuration."""
        return sorted(self._config["environments"].iterkeys())

    def get(self, name):
        """Retrieve the Environment with the given name.

        @return: The Environment, or None if one isn't found.
        """
        environment_config = self._config["environments"].get(name)
        if environment_config is not None:
            return Environment(name, environment_config)
        return None

    def get_default(self):
        """Get the default environment for this configuration.

        The default environment is either the single defined environment
        in the configuration, or the one explicitly named through the
        "default:" option in the outermost scope.

        @raise EnvironmentsConfigError: If it can't determine a default
            environment.
        """
        environments_config = self._config.get("environments")
        if len(environments_config) == 1:
            return self.get(environments_config.keys()[0])
        default = self._config.get("default")
        if default:
            if default not in environments_config:
                raise EnvironmentsConfigError(
                    "Default environment '%s' was not found: %s" %
                    (default, self._loaded_path))
            return self.get(default)
        raise EnvironmentsConfigError("There are multiple environments and no "
                                     "explicit default (set one explicitly?): "
                                     "%s" % self._loaded_path)

    def write_sample(self, path=None):
        """Write down a sample configuration file.

        @param path: An optional environment configuration file path.
            Defaults to ~/.ensemble/environments.yaml
        """
        path = self._get_path(path)
        dirname = os.path.dirname(path)
        if os.path.exists(path):
            raise FileAlreadyExists(path)
        if not os.path.exists(dirname):
            os.makedirs(dirname)

        defaults = {
            "control-bucket": "ensemble-%s" % (uuid.uuid4().hex),
            "admin-secret": "%s" % (uuid.uuid4().hex)}

        with open(path, "w") as file:
            file.write(SAMPLE_CONFIG % defaults)

    def load_or_write_sample(self):
        """Try to load the configuration, and if it doesn't work dump a sample.

        This method will try to load the environment configuration from the
        default location, and if it doesn't work, it will write down a
        sample configuration there.

        This is handy for a default initialization.
        """
        try:
            self.load()
        except FileNotFound:
            self.write_sample()
            raise EnvironmentsConfigError("No environments configured. Please "
                                          "edit: %s" % self.get_default_path(),
                                          sample_written=True)
