import os
import yaml

from ensemble.environment.config import EnvironmentsConfig, SAMPLE_CONFIG
from ensemble.environment.environment import Environment
from ensemble.environment.errors import EnvironmentsConfigError
from ensemble.errors import (
    FileNotFound, FileAlreadyExists, InvalidEnsembleHeaderValue)

from ensemble.lib.testing import TestCase


SAMPLE_ENV = """
ensemble: environments

environments:
  myfirstenv:
    type: dummy
    foo: bar
  mysecondenv:
    type: dummy
    nofoo: 1
"""


class EnvironmentsConfigTestBase(TestCase):

    def setUp(self):
        self.old_home = os.environ.get("HOME")
        self.tmp_home = self.makeDir()
        os.environ["HOME"] = self.tmp_home
        self.default_path = os.path.join(self.tmp_home,
                                         ".ensemble/environments.yaml")
        self.other_path = os.path.join(self.tmp_home,
                                       ".ensemble/other-environments.yaml")
        self.config = EnvironmentsConfig()

    def tearDown(self):
        os.environ["HOME"] = self.old_home

    def write_config(self, config_text, other_path=False):
        if other_path:
            path = self.other_path
        else:
            path = self.default_path
        os.makedirs(os.path.dirname(path))
        with open(path, "w") as file:
            file.write(config_text)


class EnvironmentsConfigTest(EnvironmentsConfigTestBase):

    def test_get_default_path(self):
        self.assertEquals(self.config.get_default_path(), self.default_path)

    def compare_config(self, config1, sample_config2):
        config1 = yaml.load(config1)
        config2 = yaml.load(
            sample_config2 % config1["environments"]["sample"])
        self.assertEqual(config1, config2)

    def test_load_with_nonexistent_default_path(self):
        """
        Raise an error if load() is called without a path and the
        default doesn't exist.
        """
        try:
            self.config.load()
        except FileNotFound, error:
            self.assertEquals(error.path, self.default_path)
        else:
            self.fail("FileNotFound not raised")

    def test_load_with_nonexistent_custom_path(self):
        """
        Raise an error if load() is called with non-existing path.
        """
        path = "/non/existent/custom/path"
        try:
            self.config.load(path)
        except FileNotFound, error:
            self.assertEquals(error.path, path)
        else:
            self.fail("FileNotFound not raised")

    def test_write_sample_environment_default_path(self):
        """
        write_sample() should write a pre-defined sample configuration file.
        """
        self.config.write_sample()
        self.assertTrue(os.path.isfile(self.default_path))
        with open(self.default_path) as file:
            self.compare_config(file.read(), SAMPLE_CONFIG)

    def test_write_sample_contains_secret_key_and_control_bucket(self):
        """
        write_sample() should write a pre-defined sample with an ec2 machine
        provider type, a unique s3 control bucket, and an admin secret key.
        """
        uuid_factory = self.mocker.replace("uuid.uuid4")
        uuid_factory().hex
        self.mocker.result("abc")
        uuid_factory().hex
        self.mocker.result("xyz")
        self.mocker.replay()

        self.config.write_sample()
        self.assertTrue(os.path.isfile(self.default_path))

        with open(self.default_path) as file:
            config = yaml.load(file.read())
            self.assertEqual(
                config["environments"]["sample"]["type"], "ec2")
            self.assertEqual(
                config["environments"]["sample"]["control-bucket"],
                "ensemble-abc")
            self.assertEqual(
                config["environments"]["sample"]["admin-secret"],
                "xyz")

    def test_write_sample_environment_with_default_path_and_existing_dir(self):
        """
        write_sample() should not fail if the config directory already exists.
        """
        os.mkdir(os.path.dirname(self.default_path))
        self.config.write_sample()
        self.assertTrue(os.path.isfile(self.default_path))
        with open(self.default_path) as file:
            self.compare_config(file.read(), SAMPLE_CONFIG)

    def test_write_sample_environment_with_custom_path(self):
        """
        write_sample() may receive an argument with a custom path.
        """
        path = os.path.join(self.tmp_home, "sample-file")
        self.config.write_sample(path)
        self.assertTrue(os.path.isfile(path))
        with open(path) as file:
            self.compare_config(file.read(), SAMPLE_CONFIG)

    def test_write_sample_wont_overwrite_existing_configuration(self):
        """
        write_sample() must never overwrite an existing file.
        """
        path = self.other_path
        os.makedirs(os.path.dirname(path))
        with open(path, "w") as file:
            file.write("previous content")
        try:
            self.config.write_sample(path)
        except FileAlreadyExists, error:
            self.assertEquals(error.path, path)
        else:
            self.fail("FileAlreadyExists not raised")

    def test_load_environment_with_bad_header(self):
        """
        The YAML file should have an "ensemble: environments" header.  If
        it doesn't, load() should raise a sensible error.
        """
        # Use a different path to ensure the error message is right.
        self.write_config("""
            ensemble: bad
            """, other_path=True)
        try:
            self.config.load(self.other_path)
        except InvalidEnsembleHeaderValue, error:
            self.assertEquals(error.path, self.other_path)
            self.assertEquals(error.expected, "environments")
            self.assertEquals(error.found, "bad")
        else:
            self.fail("InvalidEnsembleHeaderValue not raised")

    def test_load_empty_environments(self):
        """
        load() must raise an error if there are no enviroments defined
        in the configuration file.
        """
        # Use a different path to ensure the error message is right.
        self.write_config("""
            ensemble: environments

            environments:
            """, other_path=True)
        try:
            self.config.load(self.other_path)
        except EnvironmentsConfigError, error:
            self.assertEquals(str(error),
                "Environments configuration error: %s: "
                "environments: expected dict, got None"
                % self.other_path)
        else:
            self.fail("EnvironmentsConfigError not raised")

    def test_load_environments_with_wrong_type(self):
        """
        load() must raise an error if the "environments:" option in
        the YAML configuration file doesn't have a mapping underneath it.
        """
        # Use a different path to ensure the error message is right.
        self.write_config("""
            ensemble: environments

            environments:
              - list
            """, other_path=True)
        try:
            self.config.load(self.other_path)
        except EnvironmentsConfigError, error:
            self.assertEquals(str(error),
                "Environments configuration error: %s: "
                "environments: expected dict, got ['list']"
                % self.other_path)
        else:
            self.fail("EnvironmentsConfigError not raised")

    def test_wb_parse(self):
        """
        We'll have an exception, and use mocker here to test the
        implementation itself, because we don't want to repeat the
        same tests for both parse() and load(), so we'll just verify
        that one calls the other internally.
        """
        mock = self.mocker.patch(self.config)
        mock.parse(SAMPLE_ENV, self.other_path)
        self.write_config(SAMPLE_ENV, other_path=True)
        self.mocker.replay()
        self.config.load(self.other_path)

    def test_parse_errors_without_filename(self):
        """
        parse() may receive None as the file path, in which case the
        error should not mention it.
        """
        # Use a different path to ensure the error message is right.
        try:
            self.config.parse("""
                ensemble: environments

                environments:
                """)
        except EnvironmentsConfigError, error:
            self.assertEquals(str(error),
                "Environments configuration error: "
                "environments: expected dict, got None")
        else:
            self.fail("EnvironmentsConfigError not raised")

    def test_get_environment_names(self):
        """
        get_names() should return of the environments names contained
        in the configuration file.
        """
        self.write_config(SAMPLE_ENV)
        self.config.load()
        self.assertEquals(self.config.get_names(),
                          ["myfirstenv", "mysecondenv"])

    def test_get_non_existing_environment(self):
        """
        Trying to get() a non-existing configuration name should return None.
        """
        self.config.parse(SAMPLE_ENV)
        self.assertEquals(self.config.get("non-existing"), None)

    def test_load_and_get_environment(self):
        """
        get() should return an Environment instance.
        """
        self.write_config(SAMPLE_ENV)
        self.config.load()
        self.assertEquals(type(self.config.get("myfirstenv")), Environment)

    def test_load_or_write_sample_with_non_existent_config(self):
        """
        When an environment configuration does not exist, the method
        load_or_write_sample() must write down a sample configuration
        file, and raise an error to let the user know his request did
        not work, and he should edit this file.
        """
        try:
            self.config.load_or_write_sample()
        except EnvironmentsConfigError, error:
            self.assertEquals(str(error),
                "No environments configured. Please edit: %s" %
                self.default_path)
            self.assertEquals(error.sample_written, True)
            with open(self.default_path) as file:
                self.compare_config(file.read(), SAMPLE_CONFIG)
        else:
            self.fail("EnvironmentsConfigError not raised")

    def test_environment_config_error_sample_written_defaults_to_false(self):
        """
        The error raised by load_or_write_sample() has a flag to let the
        calling site know if a sample file was actually written down or not.
        It must default to false, naturally.
        """
        error = EnvironmentsConfigError("message")
        self.assertFalse(error.sample_written)

    def test_load_or_write_sample_will_load(self):
        """
        load_or_write_sample() must load the configuration file if it exists.
        """
        self.write_config(SAMPLE_ENV)
        self.config.load_or_write_sample()
        self.assertTrue(self.config.get("myfirstenv"))

    def test_get_default_with_single_environment(self):
        """
        get_default() must return the one defined environment, when it's
        indeed a single one.
        """
        config = yaml.load(SAMPLE_ENV)
        del config["environments"]["mysecondenv"]
        self.write_config(yaml.dump(config))
        self.config.load()
        env = self.config.get_default()
        self.assertEquals(env.name, "myfirstenv")

    def test_get_default_with_named_default(self):
        """
        get_default() must otherwise return the environment named
        through the "default:" option.
        """
        config = yaml.load(SAMPLE_ENV)
        config["default"] = "mysecondenv"
        self.write_config(yaml.dump(config))
        self.config.load()
        env = self.config.get_default()
        self.assertEquals(env.name, "mysecondenv")

    def test_default_is_schema_protected(self):
        """
        The schema should mention the "default:" option as a string.
        """
        config = yaml.load(SAMPLE_ENV)
        config["default"] = 1
        self.write_config(yaml.dump(config))
        error = self.assertRaises(EnvironmentsConfigError, self.config.load)
        self.assertEquals(
            str(error),
            "Environments configuration error: %s: "
            "default: expected string, got 1" % self.default_path)

    def test_get_default_with_named_but_missing_default(self):
        """
        get_default() must raise an error if the environment named through
        the "default:" option isn't found.
        """
        config = yaml.load(SAMPLE_ENV)
        config["default"] = "non-existent"
        # Use a different path to ensure the error message is right.
        self.write_config(yaml.dump(config), other_path=True)
        self.config.load(self.other_path)
        try:
            self.config.get_default()
        except EnvironmentsConfigError, error:
            self.assertEquals(str(error),
                "Default environment 'non-existent' was not found: "
                + self.other_path)
        else:
            self.fail("EnvironmentsConfigError not raised")

    def test_get_default_without_computable_default(self):
        """
        get_default() must raise an error if there are multiple defined
        environments and no explicit default was defined.
        """
        # Use a different path to ensure the error message is right.
        self.write_config(SAMPLE_ENV, other_path=True)
        self.config.load(self.other_path)
        try:
            self.config.get_default()
        except EnvironmentsConfigError, error:
            self.assertEquals(str(error),
                "There are multiple environments and no explicit default "
                "(set one explicitly?): " + self.other_path)
        else:
            self.fail("EnvironmentsConfigError not raised")

    def test_ensure_provider_types_are_set(self):
        """
        The schema should refuse to receive a configuration which
        contains a machine provider configuration without any type
        information.
        """
        config = yaml.load(SAMPLE_ENV)
        # Delete the type.
        del config["environments"]["myfirstenv"]["type"]
        self.write_config(yaml.dump(config), other_path=True)
        try:
            self.config.load(self.other_path)
        except EnvironmentsConfigError, error:
            self.assertEquals(str(error),
                "Environments configuration error: %s: "
                "environments.myfirstenv.type: required value not found"
                % self.other_path)
        else:
            self.fail("EnvironmentsConfigError not raised")
