"""
This file holds the generic errors which are sensible for several
areas of Ensemble.
"""


class EnsembleError(Exception):
    """All errors in Ensemble are subclasses of this.
 
    This error should not be raised by itself, though, since it means
    pretty much nothing.  It's useful mostly as something to catch instead.
    """


class FileNotFound(EnsembleError):
    """Raised when a file is not found, obviously! :-)

    @ivar path: Path of the directory or file which wasn't found.
    """

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "File was not found: %s" % (self.path,)


class FormulaError(EnsembleError):
    """An error occurred while processing a formula."""

    def __init__(self, path, message):
        self.path = path
        self.message = message

    def __str__(self):
        return "Error processing '%s': %s." % (self.path, self.message)


class FileAlreadyExists(EnsembleError):
    """Raised when something refuses to overwrite an existing file.

    @ivar path: Path of the directory or file which wasn't found.
    """

    def __init__(self, path):
        self.path = path

    def __str__(self):
        return "File already exists, won't overwrite: %s" % (self.path,)


class InvalidEnsembleHeaderValue(EnsembleError):
    """Raised when the ensemble: option value in the YAML file is unexpected.

    @ivar expected: The value which was actually expected.
    @ivar found: The value which was found.
    """

    def __init__(self, path, expected, found):
        self.expected = expected
        self.found = found
        self.path = path

    def __str__(self):
        return ("Expected a YAML file with an 'ensemble: %s' "
                "header, found 'ensemble: %s': %s"
                % (self.expected, self.found, self.path))


class NoConnection(EnsembleError):
    """Raised when the admin cli is unable to establish a Zookeeper connection.
    """


class ProviderError(EnsembleError):
    """Raised when an exception occurs in a provider."""
