import os
import hashlib
import zipfile
import tempfile

from ensemble.lib.filehash import compute_file_hash
from ensemble.formula.metadata import MetaData
from ensemble.formula.bundle import FormulaBundle
from ensemble.formula.base import FormulaBase


class FormulaDirectory(FormulaBase):
    """Directory that holds formula content.

    :param path: Path to formula directory

    The directory must contain the following files::

    - ``metadata.yaml``

    """

    def __init__(self, path):
        self.path = path
        self.metadata = MetaData(os.path.join(path, "metadata.yaml"))
        self._temp_bundle = None
        self._temp_bundle_file = None

    @classmethod
    def from_archive(cls, archive_path, directory_path):
        zf = zipfile.ZipFile(archive_path, 'r')
        zf.extractall(directory_path)
        return cls(directory_path)

    def make_archive(self, path):
        """Create archive of directory and write to ``path``.

        :param path: Path to archive

        - build/* - This is used for packing the formula itself and any
                    similar tasks.
        - */.*    - Hidden files are all ignored for now.  This will most
                    likely be changed into a specific ignore list (.bzr, etc)
        """

        zf = zipfile.ZipFile(path, 'w')

        for dirpath, dirnames, filenames in os.walk(self.path):
            relative_path = dirpath[len(self.path) + 1:]
            if relative_path and not self._ignore(relative_path):
                zf.write(dirpath, relative_path)
            for name in filenames:
                archive_name = os.path.join(relative_path, name)
                if not self._ignore(archive_name):
                    real_path = os.path.join(dirpath, name)
                    zf.write(real_path, archive_name)

        zf.close()

    def _ignore(self, path):
        if path == "build" or path.startswith("build/"):
            return True
        if path.startswith('.'):
            return True

    def as_bundle(self):
        if self._temp_bundle is None:
            prefix = "%s-%d.frml." % \
                     (self.metadata.name, self.metadata.revision)
            temp_file = tempfile.NamedTemporaryFile(prefix=prefix)
            self.make_archive(temp_file.name)
            self._temp_bundle = FormulaBundle(temp_file.name)
            # Attach the life time of temp_file to self:
            self._temp_bundle_file = temp_file
        return self._temp_bundle

    def compute_sha256(self):
        """
        Compute sha256, based on the bundle.
        """
        return self.as_bundle().compute_sha256()
