import os

import yaml

from ensemble.errors import (
    EnsembleError, InvalidEnsembleHeaderValue, FileNotFound)
from ensemble.lib.schema import (
    SchemaError, KeyDict, Constant, Int, UnicodeOrString)


class MetaDataError(EnsembleError):
    """Raised when an error in the info file of a formula is found."""


SCHEMA = KeyDict({
    "ensemble": Constant("formula"),
    "name": UnicodeOrString("utf-8"),
    "revision": Int(),
    "summary": UnicodeOrString("utf-8"),
    "description": UnicodeOrString("utf-8"),
    })


class MetaData(object):
    """Represents the formula info file.

    The main metadata for a formula (name, revision, etc) is maintained
    in the formula's info file.  This class is able to parse,
    validate, and provide access to data in the info file.
    """

    def __init__(self, path=None):
        self._data = {}
        self._raw_data = ""
        if path is not None:
            self.load(path)

    @property
    def name(self):
        return self._data.get("name")

    @property
    def revision(self):
        """The formula revision.

        The formula revision acts as a version, but unlike e.g. package
        versions, the formula revision is a monotonically increasing
        integer.
        """
        return self._data.get("revision")

    @property
    def summary(self):
        return self._data.get("summary")

    @property
    def description(self):
        return self._data.get("description")

    def load(self, path):
        """Load and parse the info file.

        @param path: Path of the file to load.

        Internally, this function will pass the content of the file to
        the C{parse()} method.
        """
        if not os.path.isfile(path):
            raise FileNotFound(path)
        with open(path) as f:
            self.parse(f.read(), path)

    def parse(self, content, path=None):
        """Parse the info file described by the given content.

        @param content: Content of the info file to parse.
        @param path: Optional path of the loaded file.  Used when raising
            errors.

        @raise MetaDataError: When errors are found in the info data.
        """
        self._raw_data = content
        data = yaml.load(content)
        expected_value = "formula"
        found_value = data.get("ensemble")
        if found_value != expected_value:
            raise InvalidEnsembleHeaderValue(path, expected_value, found_value)
        try:
            self._data = SCHEMA.coerce(data, [])
        except SchemaError, error:
            if path:
                path_info = " %s:" % path
            else:
                path_info = ""
            raise MetaDataError("Bad data in formula info:%s %s" %
                                (path_info, error))

    def get_raw_data(self):
        return self._raw_data
