import os
import hashlib
import inspect
import zipfile

from ensemble.lib.testing import TestCase
from ensemble.lib.filehash import compute_file_hash
from ensemble.formula.metadata import MetaData
from ensemble.formula.bundle import FormulaBundle
from ensemble.errors import FormulaError
from ensemble.formula.directory import FormulaDirectory
from ensemble.formula import tests

sample_directory = os.path.join(
    os.path.dirname(inspect.getabsfile(tests)), "repository", "dummy")


class BundleTest(TestCase):

    def setUp(self):
        directory = FormulaDirectory(sample_directory)

        # add sample directory
        self.filename = self.makeFile()
        directory.make_archive(self.filename)

    def test_initialization(self):
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.path, self.filename)

    def test_error_not_zip(self):
        filename = self.makeFile("@#$@$")
        try:
            FormulaBundle(filename)
        except FormulaError, exc:
            self.assertIn(filename, str(exc))
            self.assertIn('zip', str(exc).lower())
            return

        self.fail("Expected formula error.")

    def test_error_zip_but_doesnt_have_metadata_file(self):
        filename = self.makeFile()
        zf = zipfile.ZipFile(filename, 'w')
        zf.writestr("README.txt", "This is not a valid formula.")
        zf.close()

        try:
            FormulaBundle(filename)
        except FormulaError, exc:
            self.assertIn(filename, str(exc))
            self.assertIn('metadata.yaml', str(exc).lower())
            return

        self.fail("Expected formula error.")

    def test_info(self):
        bundle = FormulaBundle(self.filename)
        self.assertTrue(bundle.metadata is not None)
        self.assertTrue(isinstance(bundle.metadata, MetaData))
        self.assertEquals(bundle.metadata.name, "dummy")

    def test_as_bundle(self):
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.as_bundle(), bundle)

    def get_formula_sha256(self):
        return compute_file_hash(hashlib.sha256, self.filename)

    def test_compute_sha256(self):
        sha256 = self.get_formula_sha256()
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.compute_sha256(), sha256)

    def test_formula_base_inheritance(self):
        """
        get_sha256() should be implemented in the base class,
        and should use compute_sha256 to calculate the digest.
        """
        sha256 = self.get_formula_sha256()
        bundle = FormulaBundle(self.filename)
        self.assertEquals(bundle.get_sha256(), sha256)
