import os
import hashlib
import inspect
import tempfile
import gc

from ensemble.errors import FileNotFound
from ensemble.formula.metadata import MetaData
from ensemble.formula.directory import FormulaDirectory
from ensemble.formula.bundle import FormulaBundle
from ensemble.lib.filehash import compute_file_hash
from ensemble.formula import tests

from ensemble.formula.tests.test_repository import RepositoryTestBase


sample_directory = os.path.join(
    os.path.dirname(inspect.getabsfile(tests)), "repository", "dummy")


class DirectoryTest(RepositoryTestBase):

    def test_metadata_is_required(self):
        directory = self.makeDir()
        self.assertRaises(FileNotFound, FormulaDirectory, directory)

    def test_info(self):
        directory = FormulaDirectory(sample_directory)
        self.assertTrue(directory.metadata is not None)
        self.assertTrue(isinstance(directory.metadata, MetaData))
        self.assertEquals(directory.metadata.name, "dummy")

    def test_from_archive(self):
        directory = FormulaDirectory(sample_directory)
        temporary = self.makeDir()
        f = tempfile.NamedTemporaryFile(suffix=".frml")

        # make archive from sample directory
        directory.make_archive(f.name)

        # extract into temporary directory
        FormulaDirectory.from_archive(f.name, temporary)

        def relative_path(basepath, *path):
            basepath = os.path.abspath(basepath)
            abspath = os.path.abspath(os.path.join(*path))
            return abspath[len(basepath) + 1:]

        manifest = set()
        for dirpath, dirnames, filenames in os.walk(temporary):
            for filename in filenames:
                manifest.add(relative_path(temporary, dirpath, filename))
            for dirname in dirnames:
                manifest.add(relative_path(temporary, dirpath, dirname))

        self.assertEqual(
            set(manifest),
            set(("metadata.yaml", "src", "src/hello.c", "empty")))

    def test_make_archive(self):
        # make archive from sample directory
        directory = FormulaDirectory(sample_directory)
        f = tempfile.NamedTemporaryFile(suffix=".frml")
        directory.make_archive(f.name)

        # open archive in .zip-format and assert integrity
        from zipfile import ZipFile
        zf = ZipFile(f.name)
        self.assertEqual(zf.testzip(), None)

        # assert included
        included = [info.filename for info in zf.infolist()]
        self.assertEqual(
            set(included),
            set(("metadata.yaml", "empty/", "src/", "src/hello.c")))

    def test_as_bundle(self):
        directory = FormulaDirectory(self.sample_dir1)
        formula_bundle = directory.as_bundle()
        self.assertEquals(type(formula_bundle), FormulaBundle)
        self.assertEquals(formula_bundle.metadata.name, "sample")
        self.assertIn("sample-1.frml", formula_bundle.path)

    def test_as_bundle_file_lifetime(self):
        """
        The temporary bundle file created should have a life time
        equivalent to that of the directory object itself.
        """
        directory = FormulaDirectory(self.sample_dir1)
        formula_bundle = directory.as_bundle()
        gc.collect()
        self.assertTrue(os.path.isfile(formula_bundle.path))
        del directory
        gc.collect()
        self.assertFalse(os.path.isfile(formula_bundle.path))

    def test_compute_sha256(self):
        """
        Computing the sha256 of a directory will use the bundled
        formula, since the hash of the file itself is needed.
        """
        directory = FormulaDirectory(self.sample_dir1)
        sha256 = directory.compute_sha256()
        formula_bundle = directory.as_bundle()
        self.assertEquals(type(formula_bundle), FormulaBundle)
        self.assertEquals(compute_file_hash(hashlib.sha256,
                                            formula_bundle.path),
                          sha256)

    def test_as_bundle_with_relative_path(self):
        """
        Ensure that as_bundle works correctly with relative paths.
        """
        current_dir = os.getcwd()
        os.chdir(self.sample_dir2)
        self.addCleanup(os.chdir, current_dir)
        formula_dir = "../%s" % os.path.basename(self.sample_dir1)

        directory = FormulaDirectory(formula_dir)
        formula_bundle = directory.as_bundle()
        self.assertEquals(type(formula_bundle), FormulaBundle)
        self.assertEquals(formula_bundle.metadata.name, "sample")

    def test_formula_base_inheritance(self):
        """
        get_sha256() should be implemented in the base class,
        and should use compute_sha256 to calculate the digest.
        """
        directory = FormulaDirectory(self.sample_dir1)
        bundle = directory.as_bundle()
        digest = compute_file_hash(hashlib.sha256, bundle.path)
        self.assertEquals(digest, directory.get_sha256())
