# -*- encoding: utf-8 -*-

import os
import yaml
import inspect

from ensemble.lib.testing import TestCase
from ensemble.formula.metadata import MetaData, MetaDataError
from ensemble.errors import InvalidEnsembleHeaderValue, FileNotFound
from ensemble.formula import tests

sample_path = os.path.join(
    os.path.dirname(inspect.getabsfile(tests)),
    "repository", "dummy", "metadata.yaml")

sample_configuration = open(sample_path).read()


class MetaDataTest(TestCase):

    def setUp(self):
        self.metadata = MetaData()
        self.sample = sample_configuration

    def change_sample(self):
        """Return a context manager for hacking the sample data.

        This should be used follows:

            with self.change_sample() as data:
                data["some-key"] = "some-data"

        The changed sample file content will be available in self.sample
        once the context is done executing.
        """

        class HackManager(object):

            def __enter__(mgr):
                mgr.data = yaml.load(self.sample)
                return mgr.data

            def __exit__(mgr, exc_type, exc_val, exc_tb):
                self.sample = yaml.dump(mgr.data)
                return False
        return HackManager()

    def test_path_argument_loads_formula_info(self):
        info = MetaData(sample_path)
        self.assertEquals(info.name, "dummy")

    def test_check_basic_info_before_loading(self):
        """
        Attributes should be set to None before anything is loaded.
        """
        self.assertEquals(self.metadata.name, None)
        self.assertEquals(self.metadata.revision, None)
        self.assertEquals(self.metadata.summary, None)
        self.assertEquals(self.metadata.description, None)

    def test_parse_and_check_basic_info(self):
        """
        Parsing the content file should work. :-)  Basic information will
        be available as attributes of the info file.
        """
        self.metadata.parse(self.sample)
        self.assertEquals(self.metadata.name, "dummy")
        self.assertEquals(self.metadata.revision, 1)
        self.assertEquals(self.metadata.summary, u"That's a dummy formula.")
        self.assertEquals(self.metadata.description,
                          u"This is a longer description which\n"
                          u"potentially contains multiple lines.\n")

    def test_load_calls_parse(self):
        """
        We'll break the rules a little bit here and test the implementation
        itself just so that we don't have to test *everything* twice. If
        load() calls parse(), then whatever happens with parse happens
        with load().
        """
        sample = "<dummy>"
        path = self.makeFile(sample)
        mock = self.mocker.patch(self.metadata)
        mock.parse(sample, path)
        self.mocker.replay()

        self.metadata.load(path)
        # Do your magic Mocker!

    def test_needs_ensemble_header(self):
        """
        The loaded content should have a "ensemble: formula" header.
        """
        with self.change_sample() as data:
            data["ensemble"] = "foo"
        error = self.assertRaises(InvalidEnsembleHeaderValue,
                                  self.metadata.parse, self.sample)
        self.assertEquals(error.expected, "formula")
        self.assertEquals(error.found, "foo")
        self.assertEquals(error.path, None)

    def test_needs_ensemble_header_with_load(self):
        """
        Check that the exception raised by load() on header errors will
        include the correct path in the exception, so that the user can
        tell where the issue is.
        """
        with self.change_sample() as data:
            data["ensemble"] = "foo"
        filename = self.makeFile(self.sample)
        error = self.assertRaises(InvalidEnsembleHeaderValue,
                                  self.metadata.load, filename)
        self.assertEquals(error.expected, "formula")
        self.assertEquals(error.found, "foo")
        # That's why we need this test:
        self.assertEquals(error.path, filename)

    def test_schema_error_includes_path_with_load(self):
        """
        When using load(), the exception message should mention the
        path name which was attempted.
        """
        with self.change_sample() as data:
            data["revision"] = "1"
        filename = self.makeFile(self.sample)
        error = self.assertRaises(MetaDataError,
                                  self.metadata.load, filename)
        self.assertEquals(str(error),
                          "Bad data in formula info: %s: revision: "
                          "expected int, got '1'" % filename)

    def test_load_missing_file(self):
        """
        When using load(), the exception message should mention the
        path name which was attempted.
        """
        filename = self.makeFile()
        error = self.assertRaises(FileNotFound,
                                  self.metadata.load, filename)
        self.assertEquals(error.path, filename)

    def test_revision_is_int(self):
        """
        Revision numbers are *integers*. Yep, no 1.2.3 versioning.
        """
        with self.change_sample() as data:
            data["revision"] = "1"
        error = self.assertRaises(MetaDataError,
                                  self.metadata.parse, self.sample)
        self.assertEquals(str(error),
                          "Bad data in formula info: revision: "
                          "expected int, got '1'")

    def test_name_summary_and_description_are_utf8(self):
        """
        Textual fields are decoded to unicode by the schema using UTF-8.
        """
        value = u"áéíóú"
        str_value = value.encode("utf-8")
        with self.change_sample() as data:
            data["name"] = str_value
            data["summary"] = str_value
            data["description"] = str_value
        self.metadata.parse(self.sample)
        self.assertEquals(self.metadata.name, value)
        self.assertEquals(self.metadata.summary, value)
        self.assertEquals(self.metadata.description, value)

    def test_get_empty_raw_data(self):
        """Before reading anything, get_raw_data() must return an empty string.
        """
        self.assertEquals(self.metadata.get_raw_data(), "")

    def test_get_raw_data(self):
        """After reading something, it should return whatever was read."""
        self.metadata.parse(self.sample)
        self.assertEquals(self.metadata.get_raw_data(), self.sample)
