import os
import inspect
import shutil

import yaml

from ensemble.lib.testing import TestCase
from ensemble.errors import FormulaError
from ensemble.formula.directory import FormulaDirectory
from ensemble.formula.repository import LocalFormulaRepository
from ensemble.formula import tests


unbundled_repository = os.path.join(
    os.path.dirname(inspect.getabsfile(tests)), "repository")


class RepositoryTestBase(TestCase):

    def setUp(self):

        self.bundled_repo_path = self.makeDir()
        self.unbundled_repo_path = self.makeDir()

        os.rmdir(self.unbundled_repo_path)

        shutil.copytree(unbundled_repository, self.unbundled_repo_path)

        self.sample_dir1 = os.path.join(self.unbundled_repo_path, "old")
        self.sample_dir2 = os.path.join(self.unbundled_repo_path, "new")


class RepositoryTest(RepositoryTestBase):

    def setUp(self):
        super(RepositoryTest, self).setUp()

        # bundle sample formulas
        FormulaDirectory(self.sample_dir1).make_archive(
            os.path.join(self.bundled_repo_path, "old.frml"))
        FormulaDirectory(self.sample_dir2).make_archive(
            os.path.join(self.bundled_repo_path, "new.frml"))

        # define repository objects
        self.repository1 = LocalFormulaRepository(self.unbundled_repo_path)
        self.repository2 = LocalFormulaRepository(self.bundled_repo_path)

    def test_find_formula_by_name_which_is_unbundled(self):
        formula = self.repository1.find("sample")
        self.assertTrue(formula)
        self.assertTrue(formula.metadata)

    def test_find_formula_ignores_unknown_files(self):
        self.makeFile(
            "Foobar",
            path=os.path.join(self.repository1.path, "zebra"))
        self.assertRaises(FormulaError, self.repository1.find, "zebra")

    def test_find_formula_ignores_unknown_directories(self):
        self.makeDir(
            path=os.path.join(self.repository1.path, "zebra"))
        self.assertRaises(FormulaError, self.repository1.find, "zebra")

    def test_find_formula_by_name_which_is_bundled(self):
        formula = self.repository2.find("sample")
        self.assertTrue(formula)
        self.assertTrue(formula.metadata)

    def test_find_formula_by_name_fails(self):
        exc = self.assertRaises(FormulaError, self.repository1.find, "missing")
        self.assertIn("missing", str(exc))
        self.assertIn(self.unbundled_repo_path, str(exc),
                      "Repository path not found in the error message.")

    def test_find_formula_with_multiple_versions_returns_latest(self):
        # Copy the repository out of the codebase, so that we can hack it.
        formula = self.repository1.find("sample")
        self.assertEquals(formula.metadata.revision, 2)

        # Invert the "latest" logic, to ensure it's not just a coincidence.
        file = open(
            os.path.join(self.repository1.path, "old/metadata.yaml"), "rw+")
        data = yaml.load(file.read())
        data["revision"] = 3
        file.seek(0)
        file.write(yaml.dump(data))
        file.close()

        formula = self.repository1.find("sample")
        self.assertEquals(formula.metadata.revision, 3)
