"""
Functional tests for secure zookeeper connections using an ssh forwarded port.

Requirements for functional test
 - sshd running on localhost
 - zookeeper running on localhost, listening on port 2181
 - user can log into localhost via key authentication
 - ~/.ssh/id_dsa exists and is configured as an authorized key
"""

import os
import pwd
import zookeeper

from txzookeeper.client import ConnectionTimeoutException

from ensemble.lib.testing import TestCase
from ensemble.state.sshclient import SSHClient


class ConnectionTest(TestCase):

    def setUp(self):
        super(ConnectionTest, self).setUp()
        self.username = pwd.getpwuid(os.getuid())[0]
        self.log = self.capture_logging("ensemble.storage.connection")
        self.old_user_name = SSHClient.remote_user
        SSHClient.remote_user = self.username
        zookeeper.set_debug_level(0)

    def tearDown(self):
        super(ConnectionTest, self).tearDown()
        zookeeper.set_debug_level(zookeeper.LOG_LEVEL_DEBUG)
        SSHClient.remote_user = self.old_user_name

    def test_connect(self):
        """
        Forwarding a port spawns an ssh process with port forwarding arguments.
        """
        connect_deferred = SSHClient().connect(
            "localhost:2181", timeout=10)

        def validate_connected(client):
            self.assertTrue(client.connected)
            client.close()

        connect_deferred.addCallback(validate_connected)
        return connect_deferred

    def xtest_timeout(self):
        """
        if a connection can not be made before a timeout period, an exception
        is raised.

        with the sshclient layer, tihs test no longer returns a failure..
        and its hard to cleanup the process tunnel..
        """
        SSHClient.remote_user = "rabbit"
        connect_deferred = SSHClient().connect(
            "foobar.example.com:2181", timeout=1)
        self.failUnlessFailure(connect_deferred, ConnectionTimeoutException)
        return connect_deferred
