"""
Functional test for EC2 Provider.

Requirements for functional test
 - valid amazon credentials present in the environment as
   AWS_SECRET_ACCESS_KEY, AWS_ACCESS_KEY_ID
 - an ssh key (id_dsa, id_rsa, identity) present in ~/.ssh
 - if the key has a password an ssh-agent must be running with
   SSH_AGENT_PID and SSH_AUTH_SOCK set in the environment.

These tests may take several minutes for each test.
"""

import os
import pwd
import sys
import zookeeper

from twisted.internet.defer import inlineCallbacks, Deferred

from ensemble.errors import FileNotFound
from ensemble.providers.ec2 import MachineProvider
from ensemble.providers.ec2.utils import wait_for_startup
from ensemble.state.sshclient import SSHClient

from ensemble.lib.testing import TestCase


class EC2ProviderFunctionalTest(TestCase):

    def setUp(self):
        super(EC2ProviderFunctionalTest, self).setUp()
        self.username = pwd.getpwuid(os.getuid())[0]
        self.log = self.capture_logging("ensemble")
        #zookeeper.set_debug_level(0)
        self.provider = MachineProvider("ec2-functional",
            {"control-bucket": "ensemble-test-%s" % (self.username)})


class EC2MachineTest(EC2ProviderFunctionalTest):

    def _filter_instances(self, instances):
        provider_instances = []
        for i in instances:
            if i.instance_state not in ("running", "pending"):
                continue
            if not self.provider.environment_name in i.reservation.groups:
                continue
            provider_instances.append(i)
        return provider_instances

    @inlineCallbacks
    def test_shutdown(self):
        """
        Shutting down the provider, terminates all instances associated to
        the provider instance.
        """
        running = yield self.provider.ec2.describe_instances()
        running_prior = set(self._filter_instances(running))

        result = yield self.provider.shutdown()

        running = yield self.provider.ec2.describe_instances()
        running_now = set(self._filter_instances(running))

        if result:
            shutdown = running_prior-running_now
            self.assertEqual(len(result), len(shutdown))
            for i in shutdown:
                self.assertTrue(i.dns_name in result)

    @inlineCallbacks
    def test_bootstrap_and_connect(self):
        """
        Launching a bootstrap instance, creates an ec2 instance with
        a zookeeper server running on it. This test may take up to 7m
        """

        machines = yield self.provider.bootstrap()
        instances = yield wait_for_startup(self.provider.ec2, machines)

        test_complete = Deferred()

        def verify_running():
            sys.stderr.write("running; ")

            def validate_connected(client):
                self.assertTrue(client.connected)
                sys.stderr.write("connected.")
                client.close()
                test_complete.callback(None)

            def validate_failure(failure):
                sys.stderr.write("failure.")
                test_complete.errback(failure)

            server = "%s:2181" % instances[0].dns_name
            client_deferred = SSHClient().connect(server, timeout=30)
            client_deferred.addCallback(validate_connected)
            client_deferred.addErrback(validate_failure)

        from twisted.internet import reactor
        reactor.callLater(30, verify_running)

        yield test_complete

    # set the timeout to something more reasonable for bootstraping
    test_bootstrap_and_connect.timeout = 300


class EC2StorageTest(EC2ProviderFunctionalTest):

    def setUp(self):
        super(EC2StorageTest, self).setUp()
        self.s3 = self.provider.s3
        self.storage = self.provider.get_file_storage()
        self.control_bucket = self.provider.config.get("control-bucket")
        return self.s3.create_bucket(self.control_bucket)

    @inlineCallbacks
    def tearDown(self):
        listing = yield self.s3.get_bucket(self.control_bucket)
        for ob in listing.contents:
            yield self.s3.delete_object(self.control_bucket, ob.key)
        yield self.s3.delete_bucket(self.control_bucket)

    @inlineCallbacks
    def test_put_object(self):
        content = "snakes eat rubies"
        file_path = self.makeFile(content)
        yield self.storage.put("files/reptile.txt", file_path)
        s3_content = yield self.s3.get_object(
            self.control_bucket, "files/reptile.txt")
        self.assertEqual(content, s3_content)

    @inlineCallbacks
    def test_get_object(self):
        content = "snakes eat rubies"
        file_path = self.makeFile(content)
        yield self.storage.put("files/reptile.txt", file_path)
        local_path = yield self.storage.get("files/reptile.txt")
        s3_content = local_path.read()
        self.assertEqual(content, s3_content)

    def test_get_object_nonexistant(self):
        remote_path = "files/reptile.txt"
        d = self.storage.get(remote_path)
        self.failUnlessFailure(d, FileNotFound)

        def validate_error_message(result):
            self.assertEqual(
                result.path, "s3://%s/%s"%(self.control_bucket, remote_path))

        d.addCallback(validate_error_message)
        return d
