from txaws.service import AWSServiceRegion, REGION_US

from connect import EC2Connect
from files import FileStorage
from iterate import EC2MachineIteration
from launch import EC2LaunchMachine, EC2Bootstrap
from shutdown import EC2Shutdown, EC2ShutdownMachine
from state import EC2SaveState, EC2LoadState


class MachineProvider(object):

    def __init__(self, environment_name, config):
        self.environment_name = environment_name
        self.config = config
        self._service = AWSServiceRegion(
            access_key = config.get("access_key", ""),
            secret_key = config.get("secret_key", ""),
            region = config.get("region", REGION_US),
            ec2_uri = config.get("ec2_uri", ""),
            s3_uri = config.get("s3_uri", ""))
        self.s3 = self._service.get_s3_client()
        self.ec2 = self._service.get_ec2_client()

    def connect(self):
        """
        Connect to the zookeeper ensemble running in the machine provider.

        returns an open C{txzookeeper.client.ZookeeperClient} and a
        C{ensemble.storage.connection.TunnelProtocol}
        """
        connect = EC2Connect(self)
        return connect.run()

    def get_file_storage(self):
        """Retrieve the provider C{FileStorage} abstraction."""
        file_storage = FileStorage(self.s3, self.config["control-bucket"])
        return file_storage

    def list_machines(self):
        """
        List all machines running in the provider.
        """
        iteration = EC2MachineIteration(self)
        return iteration.run()

    def start_machine(self):
        """
        Start a machine in the provider.
        """
        launch = EC2LaunchMachine(self)
        return launch.run()

    def shutdown_machine(self, machine):
        """
        Stop a machine in the provider.
        """
        shutdown = EC2ShutdownMachine(self)
        return shutdown.run(machine)

    def bootstrap(self):
        """
        Bootstrap an ensemble server in the provider.
        """
        bootstrap = EC2Bootstrap(self)
        return bootstrap.run()

    def shutdown(self):
        """
        Shutdown all machine resources associated to this provider instance.
        """
        shutdown = EC2Shutdown(self)
        return shutdown.run()

    def save_state(self, state):
        """
        Save state to the provider.

        @param state
        @type dict
        """
        serializer = EC2SaveState(self)
        return serializer.run(state)

    def load_state(self):
        """
        Load state from the provider, returns a dictionary.
        """
        deserializer = EC2LoadState(self)
        return deserializer.run()
