"""
Ec2 Provider File Storage on S3
"""
import os
import tempfile

from twisted.internet.defer import fail
from txaws.s3.exception import S3Error

from ensemble.errors import FileNotFound


class FileStorage(object):

    def __init__(self, s3, control_bucket):
        self.s3 = s3
        self.control_bucket = control_bucket

    def get(self, name):
        """
        If the file exists in the storage, this method will download the file
        and return a temporary file. If it doesn't exist then a FileNotFound
        failure is returned.

        @return A deferred with path to the locally stored file.
        @type Deferred
        """
        # for now we do the simplest thing and just fetch the file
        # in a single call, s3 limits this to some mb, as we grow
        # to have formulas that might grow to this size, we can
        # revisit fetching in batches (head request for size) and
        # streaming to disk.

        d = self.s3.get_object(self.control_bucket, name)

        def on_content_retrieved(result):
            # store the file to a tempfile and return
            fh = tempfile.NamedTemporaryFile()
            fh.write(result)
            fh.seek(0)
            return fh

        def on_no_file(failure):
            # Trap file not found errors and wrap them in an application error.
            failure.trap(S3Error)
            # ensure its an invalid key error, else reraise the original
            if failure.value.get_error_code() != "NoSuchKey":
                return failure
            return fail(FileNotFound("s3://%s/%s" % (
                self.control_bucket, name)))

        d.addCallback(on_content_retrieved)
        d.addErrback(on_no_file)

        return d

    def put(self, remote_path, local_path):
        """
        Transfer and store at file_path under the given name. Returns True
        if successful .

        @return A deferred with a boolean value
        @type Deferred
        """
        if not os.path.isfile(local_path):
            return fail(FileNotFound(local_path))
        return self.s3.put_object(
            self.control_bucket, remote_path, open(local_path).read())
