from twisted.internet.defer import inlineCallbacks, returnValue
from ensemble.providers import common
from ensemble.machine import ProviderMachine
from connect import EC2ZookeeperNodeAware
from operation import EC2Operation
from utils import log, get_machine_options


DEFAULT_AVAILABILITY_ZONE = "us-east-1b"

#
# installed on any node, persists till we have packages.
INSTALL_SCRIPTS = [
    "sudo mkdir -p /usr/lib/ensemble",
    "sudo touch /usr/lib/ensemble/bootstrap.txt",
    # hack because upstart will execute in parallel with packages
    # and we want them to be done.
    'while ! dpkg -s zookeeperd | grep "Status:.*installed"; do sleep 1; done',
    'while ! dpkg -s bzr | grep "Status:.*installed"; do sleep 1; done',
    "sudo /usr/bin/bzr co lp:ensemble /usr/lib/ensemble/ensemble",
    "cd /usr/lib/ensemble/ensemble && sudo python setup.py develop",
    "sudo /usr/bin/bzr co lp:txzookeeper /usr/lib/ensemble/txzookeeper",
    "cd /usr/lib/ensemble/txzookeeper && sudo python setup.py develop",
    "sudo /bin/bash -c 'echo done > /usr/lib/ensemble/bootstrap.txt'"]


class EC2LaunchMachine(EC2Operation):

    def run(self):
        """
        Launch a machine in the EC2 provider.

        Returns a C{ProviderMachine} instance.
        """
        launch_deferred = self.launch()
        launch_deferred.addCallback(self._on_launched)
        return launch_deferred

    def _on_launched(self, instances):
        results = []
        for i in instances:
            machine = ProviderMachine()
            machine.instance_id = i
            results.append(machine)

        return results

    def _ensure_group(self, groups):
        """
        Machines launched by ensemble are tagged with a group so they can
        be distinguished from other machines that might be running on
        an EC2 account. This group can be specified explicitly or implicitly
        defined by the environment name.

        @param security_groups: The configured security groups for a machine
        instance.
        """
        # Label the instance with the ensemble name.
        d = self.ec2.describe_security_groups()
        d.addCallback(self._on_received_groups, groups)
        return d

    def _on_received_groups(self, security_groups, groups):
        ensemble_group = "ensemble-%s" % self.provider.environment_name
        group_ids = [group.name for group in security_groups]

        # Ensure the provider group is added to the instance groups.
        if not ensemble_group in groups:
            groups.append(ensemble_group)

        # Create the provider group if doesn't exist.
        if not ensemble_group in group_ids:
            log.debug("Creating ensemble provider group %s", ensemble_group)
            group_created = self.ec2.create_security_group(
                ensemble_group,
                "Ensemble group for %s" % self.provider.environment_name)

            # Authorize SSH.
            group_created.addCallback(
                lambda x: self.ec2.authorize_security_group(
                    ensemble_group,
                    ip_protocol="tcp",
                    from_port="22", to_port="22",
                    cidr_ip="0.0.0.0/0"))

            # We need to describe the group to pickup the owner_id for auth.
            group_created.addCallback(
                lambda x: self.ec2.describe_security_groups(ensemble_group))

            # Authorize Internal ZK Traffic
            group_created.addCallback(
                lambda groups: self.ec2.authorize_security_group(
                    ensemble_group,
                    source_group_name=ensemble_group,
                    source_group_owner_id=groups.pop().owner_id))

            # We use a separate callback here, because we need to wait
            # on the authorize security group deferred to succeed before
            # we can return the groups result.
            return group_created.addCallback(lambda x: groups)
        return groups

    def get_machine_variables(self):
        return dict(
            packages = list(common.DEFAULT_PACKAGES),
            repositories = list(common.DEFAULT_REPOSITORIES),
            scripts = list(INSTALL_SCRIPTS))

    @inlineCallbacks
    def launch(self):
        """
        Launch an instance node within the machine provider environment.
        """
        # Get the machine launch config/options
        public_key = common.get_user_public_key(self.config)

        variables = self.get_machine_variables()
        machine_options = yield get_machine_options(
            self.config,
            public_key = public_key,
            **variables)

        # XXX Hardcode use of ensemble image (32b, lucid, us-east-1)
        machine_options["image_id"] = "ami-62b75d0b"

        # Ensure the ensemble security group exists and is included in the
        # machine instance launch options.
        machine_options["security_groups"] = yield self._ensure_group(
            machine_options["security_groups"])

        # Launch the instance.
        instances = yield self.ec2.run_instances(**machine_options)

        returnValue(
            [instance.instance_id for instance in instances])


class EC2Bootstrap(EC2LaunchMachine, EC2ZookeeperNodeAware):
    """Amazon EC2 C{MachineProvider} environment bootstrap."""

    def run(self):
        """
        Run any bootstrap required to get a working ensemble within the machine
        provider. Returns the corresponding ensemble instances' public dns
        names within the account.
        """
        load_deferred = self.provider.load_state()
        load_deferred.addCallback(self._verify_zookeeper_nodes)
        load_deferred.addCallback(self._on_state_loaded)
        return load_deferred

    def _on_state_loaded(self, state):
        if state:
            log.debug("Existing Ensemble instance found.")
            results = []
            for instance in state:
                machine = ProviderMachine()
                machine.instance_id = instance.instance_id
                results.append(machine)
            return results


        log.debug("Launching Ensemble bootstrap instance.")
        launch_deferred = self.launch()
        # Save the state of the ensemble instances to S3.
        launch_deferred.addCallback(self._on_bootstrap_launched)
        launch_deferred.addCallback(self._on_launched)
        return launch_deferred

    def _on_bootstrap_launched(self, instances):
        d = self.provider.save_state({"zookeeper_instances": instances})
        d.addCallback(lambda x: instances)
        return d

    def get_machine_variables(self):
        variables = super(EC2Bootstrap, self).get_machine_variables()
        variables["packages"].extend(common.BOOTSTRAP_PACKAGES)
        # Download the zookeeper formula and run it in standalone mode
        #variables["scripts"].extend([
        #    "sudo wget ",
        #    "sudo ensemble-machine-agent --standalone install %s",
        #    "sudo ensemble-machine-agent ])

        # what about connecting when present.. ie. do the install of zookeeper
        # inside of container, and then reconnect when the environment is ready
        # as signaled by a file presence.
        return variables
