import os

from twisted.internet.defer import succeed, fail
from txaws.s3.exception import S3Error

from ensemble.lib.testing import TestCase
from ensemble.errors import FileNotFound
from ensemble.providers.ec2.tests.common import EC2TestMixin


class FileStorageTestCase(EC2TestMixin, TestCase):

    def get_storage(self):
        provider = self.get_provider()
        storage = provider.get_file_storage()
        return storage

    def test_put_nonexistant_file(self):
        """
        Attempting to store a nonexistant file path results in an error.
        """
        self.mocker.replay()
        file_path = self.makeFile()
        storage = self.get_storage()
        d = storage.put("gold.txt", file_path)
        self.failUnlessFailure(d, FileNotFound)
        return d

    def test_put_file(self):
        """
        A file can be put in the storage.
        """
        content = "blah blah"
        control_bucket = self.get_config()["control-bucket"]
        self.s3.put_object(
            control_bucket, "pirates/content.txt", content)
        self.mocker.result(succeed(""))
        self.mocker.replay()

        file_path = self.makeFile(content)
        storage = self.get_storage()
        d = storage.put("pirates/content.txt", file_path)

        def validate_result(result):
            self.assertIdentical(result, "")

        d.addCallback(validate_result)
        return d

    def test_get_file(self):
        """Retrieving a file from storage returns a temporary file."""
        content = "blah blah"
        control_bucket = self.get_config()["control-bucket"]
        self.s3.get_object(
            control_bucket, "pirates/content.txt")
        self.mocker.result(succeed(content))
        self.mocker.replay()

        storage = self.get_storage()
        d = storage.get("pirates/content.txt")

        def validate_result(result):
            self.assertEqual(result.read(), content)
            self.assertTrue(result.name)
            result.close()
            self.assertFalse(os.path.exists(result.name))

        d.addCallback(validate_result)
        return d

    def test_get_file_nonexistant(self):
        """Retrieving a nonexistant file raises a file not found error."""
        control_bucket = self.get_config()["control-bucket"]
        file_name = "pirates/ship.txt"
        error = S3Error("<error/>", 404)
        error.errors = [{"Code": "NoSuchKey"}]
        self.s3.get_object(control_bucket, file_name)
        self.mocker.result(fail(error))
        self.mocker.replay()

        storage = self.get_storage()
        d = storage.get(file_name)
        self.failUnlessFailure(d, FileNotFound)

        def validate_error_message(result):
            self.assertEqual(
                result.path, "s3://%s/%s"%(control_bucket, file_name))
        d.addCallback(validate_error_message)
        return d

    def test_get_file_error(self):
        """
        An unexpected error from s3 on file retrieval is exposed via the api.
        """
        control_bucket = self.get_config()["control-bucket"]
        file_name = "pirates/ship.txt"
        self.s3.get_object(
            control_bucket, file_name)
        self.mocker.result(fail(S3Error("<error/>", 503)))
        self.mocker.replay()

        storage = self.get_storage()
        d = storage.get(file_name)
        self.failUnlessFailure(d, S3Error)
        return d
