
from yaml import dump, load

from twisted.internet.defer import succeed
from txaws.ec2.model import Instance, SecurityGroup

from ensemble.machine import ProviderMachine
from ensemble.providers.ec2.launch import EC2Bootstrap
from ensemble.providers.common import (
    DEFAULT_REPOSITORIES, DEFAULT_PACKAGES, BOOTSTRAP_PACKAGES)

from ensemble.lib.testing import TestCase
from ensemble.lib.mocker import MATCH

from .common import EC2TestMixin


MATCH_AMI = MATCH(lambda image_id: image_id.startswith("ami-"))


class MachineLaunchMixin(object):

    def _mock_launch_utils(self):
        get_public_key = self.mocker.replace(
            "ensemble.providers.common.get_user_public_key")

        def match_config(arg):
            return isinstance(arg, dict)

        get_public_key(MATCH(match_config))
        self.mocker.result("zebra")
        get_ami = self.mocker.replace(
            "ensemble.providers.ec2.utils.get_current_ami")
        get_ami()
        self.mocker.result(succeed("ami-714ba518"))

    def _mock_create_group(self):
        group_name = "ensemble-%s" % self.env_name
        self.ec2.create_security_group(
            group_name, "Ensemble group for %s" % self.env_name)
        self.mocker.result(succeed(True))

        self.ec2.authorize_security_group(
            group_name, ip_protocol="tcp", from_port="22",
            to_port="22", cidr_ip="0.0.0.0/0")
        self.mocker.result(succeed([self.env_name]))

        self.ec2.describe_security_groups(group_name)
        self.mocker.result(succeed(
            [SecurityGroup(group_name, "", owner_id="123")]))

        self.ec2.authorize_security_group(
            group_name, source_group_name=group_name,
            source_group_owner_id="123")
        self.mocker.result(succeed(True))


class EC2MachineLaunchTest(EC2TestMixin, MachineLaunchMixin, TestCase):

    def test_provider_launch(self):
        """
        The provider can be used to launch a machine with a minimal set of
        required packages, repositories, and and security groups.
        """
        instance = Instance("i-foobar", "running", dns_name="x1.example.com")

        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()

        def verify_user_data(data):
            lines = data.split("\n")
            self.assertEqual(lines.pop(0), "#cloud-config")
            config = load("\n".join(lines))
            repos = [dict(source=r) for r in DEFAULT_REPOSITORIES]
            self.assertEqual(config["apt_sources"], repos)
            self.assertEqual(config["packages"], DEFAULT_PACKAGES)
            return True

        self.ec2.run_instances(
            image_id=MATCH_AMI,
            instance_type="m1.small",
            max_count=1,
            min_count=1,
            security_groups=["%s-%s" % ("ensemble", self.env_name)],
            user_data=MATCH(verify_user_data))

        self.mocker.result(succeed([instance]))
        self.mocker.replay()

        def verify_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, ProviderMachine))
            self.assertEqual(result.instance_id, instance.instance_id)

        provider = self.get_provider()
        d = provider.start_machine()
        d.addCallback(verify_result)
        return d


class EC2BootstrapTest(EC2TestMixin, MachineLaunchMixin, TestCase):

    def _mock_save(self):
        """
        Mock saving bootstrap instances to S3.
        """
        self.s3.get_bucket(self.env_name)
        self.mocker.result(succeed(True))

        def match_string(data):
            return isinstance(data, str)

        self.s3.put_object(
            self.env_name, "provider-state",
            MATCH(match_string),
            content_type="text/plain")
        self.mocker.result(succeed(True))

    def _mock_launch(self):
        """Mock launching a bootstrap machine on ec2."""

        def verify_user_data(data):
            lines = data.split("\n")
            self.assertEqual(lines.pop(0), "#cloud-config")
            config = load("\n".join(lines))
            repos = [dict(source=r) for r in DEFAULT_REPOSITORIES]
            self.assertEqual(config["apt_sources"], repos)
            self.assertEqual(
                config["packages"],
                list(DEFAULT_PACKAGES) + BOOTSTRAP_PACKAGES)
            return True

        self.ec2.run_instances(
            image_id=MATCH_AMI,
            instance_type="m1.small",
            max_count=1,
            min_count=1,
            security_groups=["ensemble-%s" % self.env_name],
            user_data=MATCH(verify_user_data))

    def test_launch_bootstrap(self):
        """
        The provider bootstrap can launch a bootstrap/zookeeper machine.
        """
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([]))
        self._mock_create_group()
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed([]))
        self._mock_save()
        self.mocker.replay()

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        return bootstrap.run()

    def test_launch_bootstrap_existing_provider_group(self):
        """
        When launching a bootstrap instance the provider will use an existing
        provider instance group.
        """
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed([]))
        self._mock_save()
        self.mocker.replay()

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        return bootstrap.run()

    def test_run_with_loaded_state(self):
        """
        If the provider bootstrap is run when there is already a running
        bootstrap instance, it will just return the existing instance.
        """
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]
        state = dump({
            "zookeeper_instances":
            [i.instance_id for i in instances]})
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(state))
        self.ec2.describe_instances(*[i.instance_id for i in instances])
        self.mocker.result(succeed(instances))
        self.mocker.replay()

        def validate_result(result):
            self.assertTrue(result)
            machine = result.pop()
            self.assertTrue(isinstance(machine, ProviderMachine))
            self.assertEqual(machine.instance_id, instances[0].instance_id)

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        d = bootstrap.run()
        d.addCallback(validate_result)
        return d

    def test_run_with_launch(self):
        """
        The provider bootstrap will launch an instance when run if there
        is no existing instance.
        """
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]

        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed(instances))
        self._mock_save()
        self.mocker.replay()

        def validate_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, ProviderMachine))
            self.assertEqual(result.instance_id, instances[0].instance_id)

        provider = self.get_provider()
        bootstrap = EC2Bootstrap(provider)
        d = bootstrap.run()
        d.addCallback(validate_result)
        return d

    def test_provider_bootstrap(self):
        """
        The ec2 provider facade instantiates and runs the bootstrap.
        """
        instances = [
            Instance("i-foobar", "running", dns_name="x1.example.com")]
        self.s3.get_object(self.env_name, "provider-state")
        self.mocker.result(succeed(None))
        self.ec2.describe_security_groups()
        self.mocker.result(succeed([
            SecurityGroup("ensemble-%s" % self.env_name, "")]))
        self._mock_launch_utils()
        self._mock_launch()
        self.mocker.result(succeed(instances))
        self._mock_save()
        self.mocker.replay()

        def validate_result(result):
            self.assertEqual(len(result), 1)
            result = result.pop()
            self.assertTrue(isinstance(result, ProviderMachine))
            self.assertEqual(result.instance_id, instances[0].instance_id)

        provider = self.get_provider()
        d = provider.bootstrap()
        d.addCallback(validate_result)
        return d
