from txaws.ec2.model import Instance, Reservation
from twisted.internet.defer import succeed

from ensemble.lib.testing import TestCase
from ensemble.providers.ec2.tests.common import EC2TestMixin

from ensemble.errors import ProviderError
from ensemble.machine import ProviderMachine
from ensemble.providers.ec2.shutdown import EC2Shutdown


class EC2ShutdownMachineTest(EC2TestMixin, TestCase):

    def test_shutdown(self):
        """
        Invoking the provider shutdown method, will issue an ec2 terminate
        call to all intsances associated to the provider instance that are
        currently running.
        """
        group_name = "ensemble-%s" % self.env_name
        terminate_instance = Instance(
            "i-foobar", "running", dns_name="x1.example.com",
            reservation=Reservation("x-1", "u-1", groups=[group_name]))
        self.ec2.describe_instances(*[terminate_instance.instance_id])
        self.mocker.result(succeed([terminate_instance]))
        self.ec2.terminate_instances(terminate_instance.instance_id)
        self.mocker.result(succeed(terminate_instance))
        self.mocker.replay()

        machine = ProviderMachine()
        machine.instance_id = "i-foobar"

        provider = self.get_provider()
        d = provider.shutdown_machine(machine)

        def validate_result(result):
            self.assertEqual(machine.instance_id, result[0].instance_id)

        d.addCallback(validate_result)
        return d

    def test_shutdown_invalid_group(self):
        """
        Attempting to shutdown a machine that does not belong to this
        provider instance raises an exception.
        """
        terminate_instance = Instance(
            "i-foobar", "running", dns_name="x1.example.com",
            reservation=Reservation("x-1", "u-1", groups=["abc"]))
        self.ec2.describe_instances(*[terminate_instance.instance_id])
        self.mocker.result(succeed([terminate_instance]))
        self.mocker.replay()

        machine = ProviderMachine()
        machine.instance_id = "i-foobar"

        provider = self.get_provider()
        d = provider.shutdown_machine(machine)

        self.failUnlessFailure(d, ProviderError)
        return d

    def test_shutdown_invalid_machine(self):
        """
        Attempting to shutdown a machine that from a different provider
        type will raise a syntaxerror.
        """
        self.mocker.replay()
        machine = ProviderMachine()
        provider = self.get_provider()
        d = provider.shutdown_machine(machine)
        self.failUnlessFailure(d, SyntaxError)
        return d


class EC2ShutdownTest(EC2TestMixin, TestCase):

    def test_shutdown(self):
        """
        The shutdown operation terminates all running and pending
        instances associated to the C{MachineProvider} instance.
        """
        group_name = "ensemble-%s" % self.env_name
        terminate_instance = Instance(
                "i-foobar", "running", dns_name="x1.example.com",
                reservation=Reservation("x-1", "u-1", groups=[group_name]))
        instances = [
            terminate_instance,
            Instance(
                "i-foobar2", "terminated", dns_name="x1.example.com",
                reservation=Reservation("x-2", "u-1",
                                        groups=[group_name])),
            Instance(
                "i-foobar3", "pending", dns_name="x1.example.com",
                reservation=Reservation("x-3", "u-1"))]

        self.ec2.describe_instances()
        self.mocker.result(succeed(instances))
        self.ec2.terminate_instances(terminate_instance.instance_id)
        self.mocker.result(succeed(terminate_instance))
        self.mocker.replay()

        provider = self.get_provider()
        shutdown = EC2Shutdown(provider)
        d = shutdown.run()

        def validate_result(result):
            self.assertEqual(result[0].instance_id,
                             terminate_instance.instance_id)

        d.addCallback(validate_result)
        return d

    def test_shutdown_no_instances(self):
        """
        If there are no instances to shutdown, running the shutdown
        operation returns None.
        """
        self.ec2.describe_instances()
        self.mocker.result(succeed([]))
        self.mocker.replay()

        provider = self.get_provider()
        shutdown = EC2Shutdown(provider)
        d = shutdown.run()

        def validate_result(result):
            self.assertEqual(result, None)

        d.addCallback(validate_result)
        return d

    def test_provider_shutdown(self):
        """The EC2 provider facade instantiates and runs the shutdown."""
        self.ec2.describe_instances()
        self.mocker.result(succeed([]))
        self.mocker.replay()

        provider = self.get_provider()
        d = provider.shutdown()

        def validate_result(result):
            self.assertEqual(result, None)

        d.addCallback(validate_result)
        return d
