import logging
import csv
import StringIO

from twisted.web.client import getPage
from twisted.internet.defer import Deferred, succeed
from yaml import dump, CDumper as Dumper

log = logging.getLogger("ensemble.ec2")


BASE_IMAGE_URI = "http://uec-images.ubuntu.com/query/"
CURRENT_IMAGE_URI = BASE_IMAGE_URI + \
              "%(ubuntu_release_name)s/%(variant)s/%(version)s.current.txt"


def get_current_ami(ubuntu_release_name="lucid", arch="i386", ebs=True,
                    region="us-east-1", daily=False, desktop=False,
                    url_fetch=None):
    """Get the latest ami for the last release of ubuntu."""
    data = {}
    data["ubuntu_release_name"] = ubuntu_release_name
    data["version"] = daily and "daily" or "released"
    data["variant"] = desktop and "desktop" or "server"
    ebs_match = ebs and "ebs" or "instance-store"

    url = CURRENT_IMAGE_URI % data
    url_fetch = url_fetch or getPage
    d = url_fetch(url)

    def extract_ami(current_data):

        data_stream = StringIO.StringIO(current_data)
        for tokens in csv.reader(data_stream, "excel-tab"):
            if tokens[4] != ebs_match:
                continue
            if tokens[5] == arch and tokens[6] == region:
                return tokens[7]
        raise LookupError((ubuntu_release_name, arch, region,
                           data["version"], data["variant"]))

    d.addCallback(extract_ami)
    return d


def wait_for_startup(client, instances, interval=2):
    """Poll EC2 waiting for instance to transition to running state."""
    from twisted.internet import reactor

    on_started = Deferred()

    def poll_instance():
        d = client.describe_instances(*[i.instance_id for i in instances])
        d.addCallback(check_status)

    def check_status(instances):
        started = filter(lambda i: i.instance_state=="running", instances)

        if len(started) == len(instances):
            on_started.callback(instances)
        else:
            log.info("waiting for instance startup...")
            reactor.callLater(interval, poll_instance)

    reactor.callLater(interval, poll_instance)
    return on_started


def format_cloud_ini(public_key, packages=(), repositories=None, scripts=None):
    """
    Format a user-data cloud-init file. This will ensure correct package
    installation, and ssh access.

    This needs to be internal to a provider as cloud-init is not universal.
    """
    cloud_config = {
        "apt-update": True,
        "apt-upgrade": True,
        "ssh_authorized_keys": [public_key],
        "packages": []}

    if packages:
        cloud_config["packages"].extend(packages)

    if repositories:
        sources = [dict(source=r) for r in repositories]
        cloud_config["apt_sources"] = sources

    if scripts:
        cloud_config["runcmd"] = scripts

    output = dump(cloud_config, Dumper=Dumper)
    output = "#cloud-config\n%s"%(output)
    return output


def get_machine_options(
    config, public_key, packages=None, repositories=None, scripts=None):
    """
    Given an ensemble environment configuration, extract the proper machine
    information to be used for launching the bootstrap node.
    """
    user_data = format_cloud_ini(
        public_key = public_key,
        packages = packages,
        repositories = repositories,
        scripts = scripts)

    instance_type = config.get("instance_type", "m1.small")
    image_id = config.get("image_id")
    if image_id is None:
        d = get_current_ami()
    else:
        d = succeed(image_id)

    def on_machine_image(current_image_id):
        return {"image_id": current_image_id,
                "min_count": 1,
                "max_count": 1,
                "instance_type": instance_type,
                "user_data": user_data,
                "security_groups": []}
    d.addCallback(on_machine_image)
    return d
