from twisted.internet.defer import (
    inlineCallbacks, returnValue, succeed)

from zookeeper import NoNodeException

from ensemble.formula.metadata import MetaData
from ensemble.state.errors import FormulaStateNotFound


class FormulaStateManager(object):
    """Manages the state of formulas in an environment."""

    def __init__(self, client):
        self._client = client

    @inlineCallbacks
    def add_formula_st(self, namespace, formula):
        """Register metadata about the provided Formula.
        
        @param namespace: The namespace this formula will be
            registered under.
        """
        metadata_content = formula.metadata.get_raw_data()
        formula_st = FormulaState(self._client, namespace,
                                  formula.metadata.name,
                                  formula.metadata.revision,
                                  metadata_content)

        # TODO This will have to handle replacements in the future
        #      (see bug #619280).
        path = formula_st.zk_path
        yield self._client.create(path, metadata_content)
        yield self._client.create(path + "/sha256", formula.get_sha256())
        returnValue(formula_st)

    @inlineCallbacks
    def get_formula_st(self, formula_id):
        """Retrieve a FormulaState for the named formula."""
        try:
            metadata, stat = yield self._client.get("/formulas/" + formula_id)
        except NoNodeException:
            raise FormulaStateNotFound(formula_id)
        rest, revision = formula_id.rsplit("-", 1)
        namespace, name = rest.split(":", 1)
        formula_st = FormulaState(self._client, namespace,
                                  name, int(revision), metadata)
        returnValue(formula_st)


class FormulaState(object):
    """State of a formula registered in an environment."""

    def __init__(self, client, namespace, name, revision,
                 metadata_content):
        self._client = client
        self._namespace = namespace
        self._name = name
        self._revision = revision

        self._metadata = MetaData()
        self._metadata.parse(metadata_content)

        # Just a health check:
        assert self._metadata.name == name
        assert self._metadata.revision == revision

    @property
    def namespace(self):
        return self._namespace

    @property
    def name(self):
        return self._name

    @property
    def revision(self):
        return self._revision

    @property
    def id(self):
        return "%s:%s-%s" % \
               (self._namespace, self._name, self._revision)

    @property
    def zk_path(self):
        """Return the path within zookeeper.

        This attribute should not be used outside of the .state
        package or for debugging.
        """
        return "/formulas/%s" % self.id

    def get_metadata(self):
        """Return a deferred MetaData."""
        # The implementation may look strange here, but the idea is
        # sane: we have immediate access to the metadata at this
        # point only because of the current ZK organization.  Let's
        # force async use so that we can change the details freely.
        return succeed(self._metadata)

    @inlineCallbacks
    def get_sha256(self):
        sha256, stat = yield self._client.get(self.zk_path + "/sha256")
        returnValue(sha256)

