import yaml

import zookeeper

from twisted.internet.defer import (
    inlineCallbacks, returnValue)

from ensemble.state.errors import MachineStateNotFound


class MachineStateManager(object):
    """Manages the state of machines in an environment."""

    def __init__(self, client):
        self._client = client

    @inlineCallbacks
    def add_machine_st(self):
        """Create a new machine state.

        @return: MachineState for the created machine.
        """
        path = yield self._client.create("/machines/machine-",
                                         flags=zookeeper.SEQUENCE)
        # XXX What if the system blows up right here?
        yield self._client.create(path + "/units")

        _, internal_id = path.rsplit("/", 1)
        returnValue(MachineState(self._client, internal_id))

    @inlineCallbacks
    def get_machine_st(self, machine_id):
        """Return deferred machine state with the given id.

        @return MachineState with the given id.

        @raise MachineStateNotFound if the id is not found.
        """
        if isinstance(machine_id, int):
            internal_id = "machine-%010d" % machine_id
        else:
            internal_id = machine_id
        exists = yield self._client.exists("/machines/%s" % internal_id)
        if not exists:
            raise MachineStateNotFound(machine_id)
        machine_st = MachineState(self._client, internal_id)
        returnValue(machine_st)


class MachineState(object):

    def __init__(self, client, internal_id):
        self._client = client
        self._internal_id = internal_id

    @property
    def id(self):
        """High-level id built using the sequence as an int."""
        _, sequence = self._internal_id.rsplit("-", 1)
        return int(sequence)

    @property
    def internal_id(self):
        """Machine's internal id, of the form machine-NNNNNNNNNN."""
        return self._internal_id


    @property
    def zk_path(self):
        """Return the path within zookeeper.

        This attribute should not be used outside of the .state
        package or for debugging.
        """
        return "/machines/" + self.internal_id
