import yaml

import zookeeper

from twisted.internet.defer import (
    inlineCallbacks, returnValue, succeed)

from ensemble.state.errors import (
    ServiceStateNotFound, ServiceUnitStateNotFound, MachineAlreadyAssigned)


class ServiceStateManager(object):
    """Manages the state of services in an environment."""

    def __init__(self, client):
        self._client = client

    @inlineCallbacks
    def add_service_st(self, service_name, formula_st):
        """Create a new service with the given name.

        @param service_name: Unique name of the service created.
        @param formula_state: FormulaState for the service.

        @return: ServiceState for the created service.
        """
        service_details = {"formula": formula_st.id}
        node_data = yaml.dump(service_details)
        service_st = ServiceState(self._client, service_name, node_data)

        path = service_st.zk_path
        yield self._client.create(path, node_data)
        yield self._client.create(path + "/units")

        returnValue(service_st)

    @inlineCallbacks
    def get_service_st(self, service_name):
        """Return deferred service state with the given name.

        @return ServiceState with the given name.

        @raise ServiceStateNotFound if the unit id is not found.
        """
        try:
            node_data, state = yield self._client.get("/services/" +
                                                      service_name)
        except zookeeper.NoNodeException:
            raise ServiceStateNotFound(service_name)
        service_st = ServiceState(self._client, service_name, node_data)
        returnValue(service_st)


class ServiceState(object):
    """State of a service registered in an environment.

    Each service is composed by units, and each unit represents an
    actual deployment of software to satisfy the needs defined in
    this service state.
    """

    def __init__(self, client, service_name, node_content):
        self._client = client
        self._service_name = service_name
        self._node_data = yaml.load(node_content)

    @property
    def service_name(self):
        """Name of the service represented by this state.
        """
        return self._service_name

    @property
    def zk_path(self):
        """Return the path within zookeeper.

        This attribute should not be used outside of the .state
        package or for debugging.
        """
        return "/services/" + self._service_name

    def get_formula_id(self):
        """Return the formula id this service is supposed to use.
        """
        # The implementation may look strange here, but the idea is
        # sane: we have immediate access to the data at this
        # point only because of the current ZK organization.  Let's
        # force async use so that we can change the details freely.
        return succeed(self._node_data["formula"])

    @inlineCallbacks
    def add_unit_st(self):
        """Add a new service unit to this state.

        @return: ServiceUnitState for the created unit.
        """
        path = yield self._client.create("%s/units/%s-" %
                                         (self.zk_path, self._service_name),
                                         flags=zookeeper.SEQUENCE)
        unit_id = path.rsplit("/", 1)[1]
        returnValue(ServiceUnitState(self._client,
                                     self._service_name,
                                     unit_id))

    @inlineCallbacks
    def get_unit_st(self, unit_id):
        """Return service unit state with the given id.

        @return: ServiceUnitState with the given id.

        @raise ServiceUnitStateNotFound if the unit id is not found.
        """
        if "/" in unit_id:
            service_name, sequence = unit_id.rsplit("/", 1)
            internal_id = "%s-%010d" % (service_name, int(sequence))
        else:
            internal_id = unit_id
        exists = yield self._client.exists("%s/units/%s" %
                                           (self.zk_path, internal_id))
        if not exists:
            raise ServiceUnitStateNotFound(unit_id)
        returnValue(ServiceUnitState(self._client,
                                     self._service_name,
                                     internal_id))


class ServiceUnitState(object):
    """State of a service unit registered in an environment.

    Each service is composed by units, and each unit represents an
    actual deployment of software to satisfy the needs defined in
    this service state.
    """

    def __init__(self, client, service_name, internal_id):
        self._client = client
        self._service_name = service_name
        self._internal_id = internal_id

    @property
    def service_name(self):
        """Service name for the service from this unit."""
        return self._service_name

    @property
    def internal_id(self):
        """Unit's internal id, of the form <service name>-NNNNNNNNNN."""
        return self._internal_id

    @property
    def id(self):
        """Get a nice user-oriented identifier for this unit.

        The id property is of the form <service name>-NNNNNNNNNN,
        which isn't nice for the user to be working with, so we
        reduce NN..N to the integer value, and also replace the
        dash by a slash, to avoid confusing the unit sequence
        with a formula revision (think wordpress-1 vs wordpress/1).
        """
        service_name, sequence = self._internal_id.split("-")
        return "%s/%d" % (service_name, int(sequence))

    @property
    def zk_path(self):
        """Return the path within zookeeper.

        This attribute should not be used outside of the .state
        package or for debugging.
        """
        return "/services/%s/units/%s" % \
               (self._service_name, self._internal_id)

    @inlineCallbacks
    def assign_to_machine(self, machine_st):
        try:
            yield self._client.create(self.zk_path + "/machine",
                                      machine_st.internal_id)
        except zookeeper.NodeExistsException:
            raise MachineAlreadyAssigned(self.id)
        # XXX What if the system blows up right here?
        yield self._client.create("%s/units/%s" %
                                  (machine_st.zk_path, self.internal_id))

    @inlineCallbacks
    def unassign_from_machine(self):
        machine_id, stat = yield self._client.get(self.zk_path + "/machine")
        try:
            yield self._client.delete(self.zk_path + "/machine",
                                      stat["version"])
        except zookeeper.BadVersionException:
            # XXX How to test this?
            return # Machine was replaced meanwhile. It's cool.
        # XXX What if the system blows up here?
        yield self._client.delete("/machines/%s/units/%s" %
                                  (machine_id, self.internal_id))
