"""
An SSH forwarding connection using ssh as a spawned process from the twisted
reactor.
"""
import os
import logging

from twisted.internet.protocol import ProcessProtocol
from ensemble.errors import FileNotFound


log = logging.getLogger("ensemble.state.sshforward")


def _verify_ports(*ports):
    for port in ports:
        if not isinstance(port, int):
            if not port.isdigit():
                raise SyntaxError("Port must be integer, got %s." % (port))


class TunnelProtocol(ProcessProtocol):

    def errReceived(self, data):
        # Every connection we'll get a warning about unknown host key being
        # added to /dev/null
        if data.startswith("Warning:"):
            return
        log.error("SSH tunnel error %s", data)


def forward_port(remote_user, local_port, remote_host, remote_port,
                 private_key_path=None):
    """
    Fork an ssh process to enable port forwarding from the given local host
    port to the remote host, remote port.

    @param local_port: The local port that should be bound for the forward.
    @type int

    @param remote_user: The user for login into the remote host.
    @type str

    @param remote_host: The name or ip address of the remote host.
    @type str

    @param remote_port: The remote port that is forwarded to.
    @type int

    @param private_key: The identity file that the private key should be read
                        from. If none is specified.
    """
    _verify_ports(local_port, remote_port)

    info = {"remote_user": remote_user,
            "local_port": local_port,
            "remote_host": remote_host,
            "remote_port": remote_port}

    from twisted.internet import reactor

    # additional args to disable password authentication attempts and to
    # disable known host checking.
    args = [
        "ssh",
        "-o", "PasswordAuthentication no",
        "-o", "UserKnownHostsFile /dev/null",
        "-o", "StrictHostKeyChecking no",
        "-Llocalhost:%(local_port)s:localhost:%(remote_port)s" % info,
        "%(remote_user)s@%(remote_host)s" % info]

    if private_key_path:
        private_key_path = os.path.expandvars(os.path.expanduser(
            private_key_path))
        if not os.path.exists(private_key_path):
            raise FileNotFound("Private key file not found: %r." % \
                               private_key_path)
        args.insert(1, private_key_path)
        args.insert(1, "-i")
        log.debug("Using private key from %s." % private_key_path)

    # use the existing process environment to utilize an ssh agent if present.
    log.debug("Spawning SSH process with %s." % (
        " ".join("%s=\"%s\"" % pair for pair in info.items())))

    return reactor.spawnProcess(
        TunnelProtocol(),
        "/usr/bin/ssh",
        args,
        env=os.environ)
