from twisted.internet.defer import inlineCallbacks

from ensemble.state.formula import FormulaStateManager
from ensemble.errors import FormulaStateNotFound

from ensemble.state.tests.common import StateTestBase


class FormulaStateManagerTest(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield StateTestBase.setUp(self)
        self.formula_stman = FormulaStateManager(self.client)

    @inlineCallbacks
    def test_add_formula(self):
        """
        Adding a Formula into a FormulaStateManager should register
        the formula within the Zookeeper state, according to the
        specification.
        """
        formula_st = yield self.formula_stman.add_formula_st(
            "namespace", self.formula)
        self.assertEquals(formula_st.id, "namespace:dummy-1")

        children = yield self.client.get_children("/formulas")
        self.assertEquals(children, ["namespace:dummy-1"])

        formula_path = "/formulas/namespace:dummy-1"
        content, state = yield self.client.get(formula_path)
        self.assertEquals(content, self.formula.metadata.get_raw_data())

    @inlineCallbacks
    def test_get_formula(self):
        """
        A FormulaState should be available if one get()s a formula
        that was previously added into the manager.
        """
        yield self.formula_stman.add_formula_st("namespace", self.formula)
        formula_st = yield self.formula_stman.get_formula_st("namespace:dummy-1")
        self.assertEquals(formula_st.id, "namespace:dummy-1")

    @inlineCallbacks
    def test_fromula_state_attributes(self):
        """
        Verify that the basic (invariant) attributes of the
        FormulaState are correctly in place.
        """
        yield self.formula_stman.add_formula_st("namespace", self.formula)
        formula_st = yield self.formula_stman.get_formula_st("namespace:dummy-1")
        self.assertEquals(formula_st.namespace, "namespace")
        self.assertEquals(formula_st.name, "dummy")
        self.assertEquals(formula_st.revision, 1)
        self.assertEquals(formula_st.id, "namespace:dummy-1")

    @inlineCallbacks
    def test_formula_state_metadata(self):
        """
        Check that the formula metadata was correctly saved and loaded.
        """
        yield self.formula_stman.add_formula_st("namespace", self.formula)
        formula_state = yield self.formula_stman.get_formula_st("namespace:dummy-1")
        metadata = yield formula_st.get_metadata()
        self.assertEquals(metadata.name, "dummy")
        self.assertEquals(metadata.revision, 1)

    @inlineCallbacks
    def test_get_missing_formula(self):
        """
        Trying to retrieve a formula from the state when it was
        never added should raise an error.
        """
        try:
            yield self.formula_stman.get_formula_st("namespace:dummy-1")
        except FormulaStateNotFound, e:
            self.assertEquals(e.formula_id, "namespace:dummy-1")
        else:
            self.fail("Error not raised.")

    @inlineCallbacks
    def test_get_sha256(self):
        """
        We should be able to retrieve the sha256 of a stored
        formula.
        """
        yield self.formula_stman.add_formula_st("namespace", self.formula)
        formula_st = yield self.formula_stman.get_formula_st("namespace:dummy-1")
        sha256 = yield formula_st.get_sha256()
        self.assertEquals(sha256, self.formula.get_sha256())

