import yaml

from twisted.internet.defer import inlineCallbacks

from ensemble.state.formula import FormulaStateManager
from ensemble.state.service import ServiceStateManager, ServiceUnitState
from ensemble.state.machine import MachineStateManager
from ensemble.state.errors import (
    ServiceStateNotFound, ServiceUnitStateNotFound, MachineAlreadyAssigned)

from ensemble.state.tests.common import StateTestBase


class ServiceStateManagerTest(StateTestBase):

    @inlineCallbacks
    def setUp(self):
        yield StateTestBase.setUp(self)
        self.formula_stman = FormulaStateManager(self.client)
        self.service_stman = ServiceStateManager(self.client)
        self.machine_stman = MachineStateManager(self.client)
        self.formula_st = yield self.formula_stman.add_formula_st(
            "namespace", self.formula)

    @inlineCallbacks
    def test_add_service(self):
        """
        Adding a service state should register it in zookeeper,
        including the requested formula id.
        """
        service_st = yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)
        children = yield self.client.get_children("/services")
        self.assertEquals(children, ["wordpress"])

        children = yield self.client.get_children("/services/wordpress")
        self.assertEquals(children, ["units"])

        content, stat = yield self.client.get("/services/wordpress")
        details = yaml.load(content)

        self.assertTrue(details)
        self.assertEquals(details.get("formula"), "namespace:dummy-1")

    @inlineCallbacks
    def test_get_service_and_check_attributes(self):
        """
        Getting a service state should be possible, and the service
        state identification should be available through its
        attributes.
        """
        yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)
        service_st = yield self.service_stman.get_service_st("wordpress")
        self.assertEquals(service_st.service_name, "wordpress")

    @inlineCallbacks
    def test_get_service_not_found(self):
        """
        Getting a service state which is not available should errback
        a meaningful error.
        """
        try:
            yield self.service_stman.get_service_st("wordpress")
        except ServiceStateNotFound, e:
            self.assertEquals(e.service_name, "wordpress")
        else:
            self.fail("Error not raised")

    @inlineCallbacks
    def test_get_service_formula_id(self):
        """
        The service state should make its respective formula id available.
        """
        yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)
        service_st = yield self.service_stman.get_service_st("wordpress")
        formula_id = yield service_st.get_formula_id()
        self.assertEquals(formula_id, "namespace:dummy-1")

    @inlineCallbacks
    def test_add_service_unit_and_check_attributes(self):
        """
        A service state should enable adding a new service unit
        under it, and again the unit should offer attributes allowing
        its identification.
        """
        service_st = yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)
        unit_st = yield service_st.add_unit_st()

        children = yield self.client.get_children("/services/wordpress/units")
        self.assertEquals(children, ["wordpress-0000000000"])

        self.assertEquals(unit_st.service_name, "wordpress")
        self.assertEquals(unit_st.id, "wordpress/0")
        self.assertEquals(unit_st.internal_id, "wordpress-0000000000")

    @inlineCallbacks
    def test_get_service_unit(self):
        """
        Getting back service units should be possible, using both
        the internal id and the nice id.
        """
        service_st = yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)

        yield service_st.add_unit_st()
        yield service_st.add_unit_st()

        unit_st1 = yield service_st.get_unit_st("wordpress/0")
        unit_st2 = yield service_st.get_unit_st("wordpress-0000000000")
        unit_st3 = yield service_st.get_unit_st("wordpress/1")
        unit_st4 = yield service_st.get_unit_st("wordpress-0000000001")

        self.assertEquals(unit_st1.internal_id, "wordpress-0000000000")
        self.assertEquals(unit_st2.internal_id, "wordpress-0000000000")
        self.assertEquals(unit_st3.internal_id, "wordpress-0000000001")
        self.assertEquals(unit_st4.internal_id, "wordpress-0000000001")

    @inlineCallbacks
    def test_get_service_unit_not_found(self):
        """
        Attempting to retrieve a non-existent service unit should
        result in an errback.
        """
        service_st = yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)

        try:
            yield service_st.get_unit_st("wordpress/0")
        except ServiceUnitStateNotFound, e:
            self.assertEquals(e.unit_id, "wordpress/0")
        else:
            self.fail("Error not raised")

    @inlineCallbacks
    def test_assign_unit_to_machine(self):
        service_st = yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)
        unit_st = yield service_st.add_unit_st()
        machine_st = yield self.machine_stman.add_machine_st()

        yield unit_st.assign_to_machine(machine_st)

        path = "/services/wordpress/units/wordpress-0000000000"
        content, stat = yield self.client.get(path + "/machine")
        self.assertEquals(content, "machine-0000000000")

        path = "/machines/machine-0000000000/units"
        children = yield self.client.get_children(path)
        self.assertEquals(children, ["wordpress-0000000000"])

    @inlineCallbacks
    def test_unassign_unit_from_machine(self):
        service_st = yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)
        unit_st = yield service_st.add_unit_st()
        machine_st = yield self.machine_stman.add_machine_st()

        yield unit_st.assign_to_machine(machine_st)
        yield unit_st.unassign_from_machine()

        path = "/services/wordpress/units/wordpress-0000000000"
        exists = yield self.client.exists(path + "/machine")
        self.assertFalse(exists)

        path = "/machines/machine-0000000000/units"
        children = yield self.client.get_children(path)
        self.assertEquals(children, [])


    @inlineCallbacks
    def test_assign_unit_to_machine_again_fails(self):
        """
        Trying to assign a machine to an already assigned unit
        should fail.
        """
        service_st = yield self.service_stman.add_service_st(
            "wordpress", self.formula_st)
        unit_st = yield service_st.add_unit_st()
        machine_st0 = yield self.machine_stman.add_machine_st()
        machine_st1 = yield self.machine_stman.add_machine_st()

        yield unit_st.assign_to_machine(machine_st0)

        try:
            yield unit_st.assign_to_machine(machine_st1)
        except MachineAlreadyAssigned, e:
            self.assertEquals(e.unit_id, "wordpress/0")
        else:
            self.fail("Error not raised")
