import os
import logging

from ensemble.lib.testing import TestCase
from ensemble.lib.mocker import MATCH
from ensemble.errors import FileNotFound
from ensemble.state.sshforward import forward_port, TunnelProtocol


IS_PROCESS_PROTOCOL = MATCH(lambda obj: isinstance(obj, TunnelProtocol))


class ConnectionTest(TestCase):

    def setUp(self):
        super(ConnectionTest, self).setUp()

    def test_invalid_forward_args(self):
        self.assertRaises(
            SyntaxError,
            forward_port,
            "ubuntu", "10000", "localhost", "1000a")

    def test_ssh_spawn(self):
        """
        Forwarding a port spawns an ssh process with port forwarding arguments.
        """
        from twisted.internet import reactor
        mock_reactor = self.mocker.patch(reactor)
        mock_reactor.spawnProcess(
            IS_PROCESS_PROTOCOL,
            "/usr/bin/ssh",
            ['ssh',
             "-o", "PasswordAuthentication no",
             "-o", "UserKnownHostsFile /dev/null",
             "-o", "StrictHostKeyChecking no",
             '-Llocalhost:8888:localhost:9999', 'ubuntu@remote_host'],
            env=os.environ)
        self.mocker.result(None)
        self.mocker.replay()
        result = forward_port("ubuntu", 8888, "remote_host", 9999)
        self.failUnlessIs(result, None)

    def test_forward_with_invalid_key(self):
        """
        Using an invalid private key with connect raises an FileNotFound
        exception.
        """
        self.assertRaises(
            FileNotFound,
            forward_port, "ubuntu", 2222, "remote", 2181,
            private_key_path="xyz-123_dsa")

    def test_connect_with_key(self):
        """
        A private key path can optionally be specified as an argument
        to connect in which case its passed to the ssh command line.
        """
        from twisted.internet import reactor
        file_path = self.makeFile("content")
        mock_reactor = self.mocker.patch(reactor)
        mock_reactor.spawnProcess(
            IS_PROCESS_PROTOCOL,
            "/usr/bin/ssh",
            ["ssh", "-i", file_path,
             "-o", "PasswordAuthentication no",
             "-o", "UserKnownHostsFile /dev/null",
             "-o", "StrictHostKeyChecking no",
             "-Llocalhost:22181:localhost:2181", "ubuntu@remote"],
            env=os.environ)
        self.mocker.result("the-process")
        self.mocker.replay()
        result = forward_port("ubuntu", 22181, "remote", 2181,
                              private_key_path=file_path)
        self.assertEquals(result, "the-process")

    def test_tunnel_protocol_logs_errors(self):
        """
        When using ssh to setup the tunnel, we log all stderr output, and
        it will get logged at the error level to the connection log handler.
        """
        log = self.capture_logging(level=logging.ERROR)
        protocol = TunnelProtocol()
        protocol.errReceived("test")
        self.assertEqual(log.getvalue(), "SSH tunnel error test\n")

    def test_tunnel_protocol_ignores_warnings(self):
        """
        When using ssh to setup the tunnel, we typically recieve a unknown host
        key warning message on stderr, the tunnel protocol will filter these
        messages.
        """
        log = self.capture_logging(level=logging.ERROR)
        protocol = TunnelProtocol()
        protocol.errReceived("Warning: Host key added")
        self.assertEqual(log.getvalue(), "")
